package fr.lifl.ioda ;

import java.util.WeakHashMap ;
import java.util.Iterator ;
import java.util.StringTokenizer ;
import java.io.IOException ;
import java.io.FileReader ;
import java.io.File ;
import java.io.BufferedReader ;
import java.io.StringReader ;
import org.nlogo.api.LogoList ;
import org.nlogo.api.Nobody ;
import org.nlogo.api.LogoListBuilder ;
import org.nlogo.api.DefaultCommand ;
import org.nlogo.api.DefaultReporter ;
import org.nlogo.api.Syntax ;
import org.nlogo.api.Argument ;
import org.nlogo.api.Context ;
import org.nlogo.nvm.ExtensionContext ;
import org.nlogo.api.ExtensionException ;
import org.nlogo.api.LogoException ;
import org.nlogo.agent.Agent ;
import org.nlogo.agent.Turtle ;
import org.nlogo.agent.Patch ;


/** <p>This class is part of the IODA NetLogo extension.<bR> All
 * contents &copy; 2008-2013 Sébastien PICAULT and Philippe
 * MATHIEU<br> Laboratoire d'Informatique Fondamentale de Lille
 * (LIFL), UMR CNRS 8021 <br> University Lille 1 – Cité Scientifique,
 * F-59655 Villeneuve d'Ascq Cedex, FRANCE.<br> Web Site: <a
 * href="http://www.lifl.fr/SMAC/projects/ioda">http://www.lifl.fr/SMAC/projects/ioda</a></p>
 * <p>The IODA NetLogo extension is free software: you can
 * redistribute it and/or modify it under the terms of the GNU General
 * Public License as published by the Free Software Foundation, either
 * version 3 of the License, or (at your option) any later
 * version.</p> <p>The IODA NetLogo extension is distributed in the
 * hope that it will be useful, but WITHOUT ANY WARRANTY; without even
 * the implied warranty of MERCHANTABILITY or FITNESS FOR A PARTICULAR
 * PURPOSE. See the GNU General Public License for more details.</p>
 * <p>You should have received a copy of the GNU General Public License
 * along with the IODA NetLogo extension. If not, see 
 * <a href="http://www.gnu.org/licenses">http://www.gnu.org/licenses</a>.</p>
 *
 * @author Sébastien Picault
 * @author Philippe Mathieu
 *
 * @see <a href="http://www2.lifl.fr/SMAC/projects/ioda/">The IODA Webpage</a>
 * @version IODA Extension 2.2 - Sept. 2013
 */
public class Interaction implements org.nlogo.api.ExtensionObject {    
    public static enum Category { PARALLEL, EXCLUSIVE } ;
    private static Category defaultCategory = Category.PARALLEL ;
    
    public static void setDefaultCategory(Category c) { defaultCategory = c ; }

    static WeakHashMap<String, Interaction> interactions = new WeakHashMap<String,Interaction>() ;
    private static long next = 0 ;


    public static void readFromString(String fileName, String fileContent) throws IOException, ExtensionException {
	StringReader fileIn = new StringReader(fileContent) ;
	BufferedReader in = new BufferedReader(fileIn) ;
	String line, name = null ;
	int number = 0 ;
	LogoListBuilder trigger, condition, actions ;
	trigger = new LogoListBuilder() ;
	condition = new LogoListBuilder() ;
	actions = new LogoListBuilder() ;
	StringTokenizer t ;
	boolean inter = false ;
	boolean exclusive = false ;
	boolean parallel = false ;
	while ((line = in.readLine()) != null) {
	    number++;
	    t = new StringTokenizer(line) ;
	    if (t.hasMoreTokens()) {
		String keyword = t.nextToken().toLowerCase() ;
		if (keyword.equals("exclusive")) {
		    if (inter) 
			throw new ExtensionException("In file: "+fileName+" line "+number+": EXCLUSIVE found inside an INTERACTION") ;	
		    if (!t.hasMoreTokens())
			throw new ExtensionException("In file: "+fileName+" line "+number+": EXCLUSIVE must be followed by INTERACTION") ;	
			exclusive = true ;
			keyword = t.nextToken().toLowerCase() ;
			if (!keyword.equals("interaction"))
			    throw new ExtensionException("In file: "+fileName+" line "+number+": EXCLUSIVE must be followed by INTERACTION") ;	
		} else if (keyword.equals("parallel")) {
		    if (inter) 
			throw new ExtensionException("In file: "+fileName+" line "+number+": PARALLEL found inside an INTERACTION") ;	
		    if (!t.hasMoreTokens())
			throw new ExtensionException("In file: "+fileName+" line "+number+": PARALLEL must be followed by INTERACTION") ;	
		    parallel = true ;
		    keyword = t.nextToken().toLowerCase() ;
		    if (!keyword.equals("interaction"))
			throw new ExtensionException("In file: "+fileName+" line "+number+": PARALLEL must be followed by INTERACTION") ;	
		}
		if (keyword.equals("interaction")) {
		    if (!t.hasMoreTokens())
			throw new ExtensionException("In file: "+fileName+" line "+number+": INTERACTION without a name") ;
		    if (inter)
			throw new ExtensionException("In file: "+fileName+" line "+number+": INTERACTION defined within another one") ;
		    name = t.nextToken().toLowerCase() ;
		    inter = true ;
		} else if (keyword.equals("end")) {
		    if (!inter)
			throw new ExtensionException("In file: "+fileName+" line "+number+": END found outside an INTERACTION") ;
		    Interaction i = new Interaction(name, trigger.toLogoList(), condition.toLogoList(), actions.toLogoList()) ;
		    if (exclusive)
			i.setExclusive(true) ;
		    if (parallel)
			i.setExclusive(false) ;
		    exclusive = false ;
		    parallel = false ;
		    trigger = new LogoListBuilder() ;
		    condition = new LogoListBuilder() ;
		    actions = new LogoListBuilder() ;
		    inter = false ; 
		} else if (keyword.equals("trigger")) {
		    if (!inter)
			throw new ExtensionException("In file: "+fileName+" line "+number+": TRIGGER found outside an INTERACTION") ;
		    LogoListBuilder tmp = new LogoListBuilder() ;
		    while (t.hasMoreTokens()) 
			tmp.add(t.nextToken()) ;
		    trigger.add(tmp.toLogoList()) ;
		} else if (keyword.equals("condition")) {
		    if (!inter)
			throw new ExtensionException("In file: "+fileName+" line "+number+": CONDITION found outside an INTERACTION") ;
		    LogoListBuilder tmp = new LogoListBuilder() ;
		    while (t.hasMoreTokens()) 
			tmp.add(t.nextToken()) ;
		    condition.add(tmp.toLogoList()) ;
		} else if (keyword.equals("actions")) {
		    if (!inter)
			throw new ExtensionException("In file: "+fileName+" line "+number+": ACTIONS found outside an INTERACTION") ;
		    while (t.hasMoreTokens()) 
			actions.add(t.nextToken()) ;
		} else if (Character.isLetterOrDigit(keyword.charAt(0)))
			throw new ExtensionException("In file: "+fileName+" line "+number+": unknown token " + keyword) ;
	    }	    
	}	
	// fileIn.close() ;
	if (inter)
	    throw new ExtensionException("In file: "+fileName+" end of file while parsing INTERACTION") ;
    }


    private final long id ;
    private LogoList trigger, condition, actions ;
    private final String name ;
    private Category status = defaultCategory ;
    
    public static void clearAll() { 
	interactions.clear() ; 
	next = 0 ;
    }

    public static Interaction get(String name) {
	if (interactions.containsKey(name.toLowerCase()))
	    return interactions.get(name.toLowerCase()) ;
	return null ;
    }

    public Interaction(String name) {
	this.name = name.toLowerCase() ;
	interactions.put(name, this) ;
	id = next ;
	next++ ;
    }

    public Interaction(String name, LogoList trigger, LogoList condition, LogoList actions) {
	this(name) ;
	setTrigger(trigger) ;
	setCondition(condition) ;
	setActions(actions) ;
    }

    public Interaction(LogoList aList) {
	this((String)aList.first()) ;
	if (aList.size() > 1)
	    setTrigger((LogoList)aList.get(1)) ;
	if (aList.size() > 2)
	    setCondition((LogoList)aList.get(2)) ;
	if (aList.size() > 3)
	    setActions((LogoList)aList.get(3)) ;
    }

    public void setTrigger(LogoList trigger) { this.trigger = trigger ; }
    public void setCondition(LogoList condition) { this.condition = condition ; }
    public void setActions(LogoList actions) { this.actions = actions ; }
    private void setCategory(Category c) { status = c ; }
    public void setExclusive(boolean t) { 
	if (t) setCategory(Category.EXCLUSIVE) ; 
	else  setCategory(Category.PARALLEL) ; 
    }

    public boolean isExclusive() { return status.equals(Category.EXCLUSIVE) ; }
    public String getName() { return name ; }
    public long getId() { return id ; }
    public LogoList getTrigger() { return (trigger == null?(new LogoListBuilder()).toLogoList():this.trigger) ; }
    public LogoList getCondition() { return (condition == null?(new LogoListBuilder()).toLogoList():this.condition) ; }
    public LogoList getActions() { return (actions == null?(new LogoListBuilder()).toLogoList():this.actions) ; }

    public boolean equals(Object obj) { return this == obj ; }

    public LogoList toList() {
	LogoListBuilder aList = new LogoListBuilder() ;
	aList.add(this.getName()) ;
	aList.add(this.getTrigger()) ;
	aList.add(this.getCondition()) ;
	aList.add(this.getActions()) ;
	return aList.toLogoList() ;
    }

    public String dump(boolean readable, boolean exportable, boolean reference) {
	if (exportable && reference) 
	    return ("" + id) ;
	else 
	    return (exportable?(id+": "):"") + org.nlogo.api.Dump.logoObject(this.toList(), true, exportable) ;
    }

    public String getExtensionName() { return "ioda" ; }
    public String getNLTypeName() { return "interaction" ; }

    public boolean recursivelyEqual(Object o) {
	if (!(o instanceof Interaction))
	    return false ;
	Interaction i = (Interaction) o ;
	if ((i.getTrigger().size() != this.getTrigger().size()) || 
	    (!(i.getTrigger().containsAll(this.getTrigger()) && (this.getTrigger().containsAll(i.getTrigger())))) ||
	    (i.getCondition().size() != this.getCondition().size()) ||
	    (!(i.getCondition().containsAll(this.getCondition()) && (this.getCondition().containsAll(i.getCondition())))))
	    return false ;
	return this.getActions().equals(i.getActions()) ;
    }

    public LogoList evalReporters(LogoList reporters, Agent source, Agent target) {
	LogoListBuilder result = new LogoListBuilder() ;
	if (reporters.size() == 0)
	    return result.toLogoList() ;
	
	String sourceBreed, targetBreed ;
	if (source instanceof Patch) 
	    sourceBreed = "patches" ;
	else sourceBreed = ((Turtle)source).getBreed().printName() ;
	if (target == null)
	    targetBreed = "nobody" ;
	else if (target instanceof Patch)
	    targetBreed = "patches" ;
	else targetBreed = ((Turtle)target).getBreed().printName() ;
	LogoListBuilder sourceIdentifier = new LogoListBuilder() ; 
	LogoListBuilder targetIdentifier = new LogoListBuilder() ;
	if (sourceBreed.equals("patches")) {
	    sourceIdentifier.add(new Double(((Patch)source).pxcor())) ;
	    sourceIdentifier.add(new Double(((Patch)source).pycor())) ;
	} else sourceIdentifier.add(new Double(source.id())) ;
	if (!targetBreed.equals("nobody")) {
	    if (targetBreed.equals("patches")) {
		targetIdentifier.add(new Double(((Patch)target).pxcor())) ;
		targetIdentifier.add(new Double(((Patch)target).pycor())) ;
	    } else targetIdentifier.add(new Double(target.id())) ;
	}
	
	Iterator<Object> it2 = reporters.iterator() ;
	while (it2.hasNext()) {
	    LogoListBuilder srcList, tgtList ;
	    LogoListBuilder tmp ;
	    srcList = new LogoListBuilder() ;
	    tgtList = new LogoListBuilder() ;
	    boolean emptySrcList = true ;
	    boolean emptyTgtList = true ;
	    LogoList tmp2 = (LogoList) it2.next() ;
	    Iterator<Object> it = tmp2.iterator() ;
	    while (it.hasNext()) {
		tmp = new LogoListBuilder() ;
		String reporter = (String) it.next() ;
		if (reporter.startsWith("target:")) {
		    if (!targetBreed.equals("nobody")) {
			reporter = reporter.substring(7) ;
			tmp.add(targetBreed+"::"+reporter) ;
			tmp.add(sourceIdentifier.toLogoList()) ;
			tgtList.add(tmp.toLogoList()) ;
			emptyTgtList = false ;
		    }
		} else {
		    tmp.add(sourceBreed+"::"+reporter) ;
		    tmp.add(targetIdentifier.toLogoList()) ;
		    srcList.add(tmp.toLogoList()) ;
		    emptySrcList = false ;
		}
	    }
	    LogoListBuilder result2 = new LogoListBuilder() ;
	    if (!emptySrcList) 
		result2.add(fput(sourceIdentifier.toLogoList(), srcList.toLogoList())) ;	
	    if (!emptyTgtList)
		result2.add(fput(targetIdentifier.toLogoList(), tgtList.toLogoList())) ;
	    result.add(result2.toLogoList()) ;
	}
	return result.toLogoList() ;
    }


    public LogoList performActions(LogoList commands, Agent source, Object target) {
	LogoListBuilder result = new LogoListBuilder() ;
	if (commands.size() == 0)
	    return result.toLogoList() ;
	boolean listTarget = false ;
	String sourceBreed, targetBreed ;
	if (source instanceof Patch) 
	    sourceBreed = "patches" ;
	else sourceBreed = ((Turtle)source).getBreed().printName() ;
	if (target == null)
	    targetBreed = "nobody" ;
	else if (target instanceof Patch)
	    targetBreed = "patches" ;
	else if (target instanceof Turtle)
	    targetBreed = ((Turtle)target).getBreed().printName() ;
	else if (target instanceof LogoList) {
	    listTarget = true ;
	    if (((LogoList)target).get(0) instanceof Patch)
		targetBreed = "patches" ;
	    else targetBreed = ((Turtle)((LogoList)target).get(0)).getBreed().printName() ;
	}
	else targetBreed = "nobody" ;
	LogoListBuilder sourceIdentifier = new LogoListBuilder() ;
	LogoListBuilder targetIdentifier = new LogoListBuilder() ; 
	if (sourceBreed.equals("patches")) {
	    sourceIdentifier.add(new Double(((Patch)source).pxcor())) ;
	    sourceIdentifier.add(new Double(((Patch)source).pycor())) ;
	} else sourceIdentifier.add(new Double(source.id())) ;
	if (!targetBreed.equals("nobody")) {
	    if (listTarget) {
		Iterator<Object> itt = ((LogoList)target).iterator() ;
		while (itt.hasNext()) {
		    LogoListBuilder tgtlist = new LogoListBuilder() ;
		    Object atgt = itt.next() ;
		    if (targetBreed.equals("patches")) {
			tgtlist.add(new Double(((Patch)atgt).pxcor())) ;
			tgtlist.add(new Double(((Patch)atgt).pycor())) ;
		    } else tgtlist.add(new Double(((Turtle)atgt).id())) ;
		    targetIdentifier.add(tgtlist.toLogoList()) ;		
		}
	    } else {
		if (targetBreed.equals("patches")) {
		    targetIdentifier.add(new Double(((Patch)target).pxcor())) ;
		    targetIdentifier.add(new Double(((Patch)target).pycor())) ;
		} else targetIdentifier.add(new Double(((Turtle)target).id())) ;
	    }
	}

	LogoListBuilder currentList = new LogoListBuilder() ;
	boolean emptyCurrentList = true ;
	boolean sourceCommands = true ;
	LogoListBuilder tmp ;
	Iterator<Object> it = commands.iterator() ;
	while (it.hasNext()) {
	    tmp = new LogoListBuilder() ;
	    String command = (String) it.next() ;
	    if (command.startsWith("target:")) {
		if (!targetBreed.equals("nobody")) {
		    if (sourceCommands && (!emptyCurrentList)) {
			result.add(fput(sourceIdentifier.toLogoList(), currentList.toLogoList()));
			currentList = new LogoListBuilder() ;
			emptyCurrentList = true ;
		    }
		    sourceCommands = false ;
		    command = command.substring(7) ;
		    tmp.add(targetBreed+"::"+command) ;
		    tmp.add(sourceIdentifier.toLogoList()) ;
		    currentList.add(tmp.toLogoList()) ;
		    emptyCurrentList = false ;
		}
	    } else {
		if (!sourceCommands && (!emptyCurrentList)) {
		    result.add(fput(targetIdentifier.toLogoList(), currentList.toLogoList())) ;
		    currentList = new LogoListBuilder() ;
		    emptyCurrentList = true ;
		}
		sourceCommands = true ;
		tmp.add(sourceBreed+"::"+command) ;
		tmp.add(targetIdentifier.toLogoList()) ;
		currentList.add(tmp.toLogoList()) ;
		emptyCurrentList = false ;
	    }
	}
	if (!emptyCurrentList) {
	    if (sourceCommands)
		result.add(fput(sourceIdentifier.toLogoList(), currentList.toLogoList())) ;
	    else 
		result.add(fput(targetIdentifier.toLogoList(), currentList.toLogoList())) ;
	}
	return result.toLogoList() ;
    }

    public static LogoList fput(LogoList l1, LogoList l2) {
	LogoListBuilder tmp = new LogoListBuilder() ;
	tmp.add(l1) ;
	tmp.addAll(l2) ;
	return tmp.toLogoList() ;
    }

    //----------------------------------------------------------------
    public static class LoadInteractions  extends DefaultCommand {
	public Syntax getSyntax() {
	    return Syntax.commandSyntax(new int[] {Syntax.StringType(), Syntax.StringType()}) ;
	}
	public String getAgentClassString() { return "OTPL" ; }
	public void perform(Argument args[], Context context) 
	    throws ExtensionException, LogoException {
	    try { //readFromFile(((ExtensionContext)context).attachCurrentDirectory(args[0].getString())) ; }
		readFromString(args[0].getString(), args[1].getString()) ; }
	    catch(IOException e) {
		throw new ExtensionException(e.getMessage()) ;
	    } 
	}
    }

    //----------------------------------------------------------------
    public static class SetExclusiveInteractions  extends DefaultCommand {
	public Syntax getSyntax() {
	    return Syntax.commandSyntax(new int[] {Syntax.BooleanType()}) ;
	}
	public String getAgentClassString() { return "OTPL" ; }
	public void perform(Argument args[], Context context) 
	    throws ExtensionException, LogoException {
	    if (args[0].getBoolean())
		setDefaultCategory(Category.EXCLUSIVE) ;
	    else 
		setDefaultCategory(Category.PARALLEL) ;
	}
    }

    //----------------------------------------------------------------
    public static class SetExclusive  extends DefaultCommand {
	public Syntax getSyntax() {
	    return Syntax.commandSyntax(new int[] {Syntax.WildcardType(), Syntax.BooleanType()}) ;
	}
	public String getAgentClassString() { return "OTPL" ; }
	public void perform(Argument args[], Context context) 
	    throws ExtensionException, LogoException {
	    Object arg0 = args[0].get() ;
	    if (! (arg0 instanceof Interaction))
		throw new ExtensionException("not an interaction: " + org.nlogo.api.Dump.logoObject(arg0)) ;
	    Interaction i = (Interaction) arg0 ;
	    i.setExclusive(args[1].getBoolean()) ;
	}
    }

    //----------------------------------------------------------------
    public static class IsExclusive extends DefaultReporter {
	public Syntax getSyntax() {
	    return Syntax.reporterSyntax(new int[] {Syntax.WildcardType()}, Syntax.BooleanType()) ;
	}
	public String getAgentClassString() { return "OTPL" ; }
	public Object report(Argument args[], Context context) 
	    throws ExtensionException, LogoException {
	    Object arg0 = args[0].get() ;
	    if (! (arg0 instanceof Interaction))
		throw new ExtensionException("not an interaction: " + org.nlogo.api.Dump.logoObject(arg0)) ;
	    Interaction i = (Interaction) arg0 ;
	    return i.isExclusive() ;
	}
    }    

    //----------------------------------------------------------------
    public static class InteractionName extends DefaultReporter {
	public Syntax getSyntax() {
	    return Syntax.reporterSyntax(new int[] {Syntax.WildcardType()}, Syntax.StringType()) ;
	}
	public String getAgentClassString() { return "OTPL" ; }
	public Object report(Argument args[], Context context) 
	    throws ExtensionException, LogoException {
	    Object arg0 = args[0].get() ;
	    if (! (arg0 instanceof Interaction))
		throw new ExtensionException("not an interaction: " + org.nlogo.api.Dump.logoObject(arg0)) ;
	    Interaction i = (Interaction) arg0 ;
	    return i.getName() ;
	}
    }    

    //----------------------------------------------------------------
    public static class InteractionMake extends DefaultReporter {
	public Syntax getSyntax() {
	    return Syntax.reporterSyntax(new int[] {Syntax.StringType(), Syntax.ListType(), Syntax.ListType(), Syntax.ListType()}, 
					 Syntax.WildcardType()) ;
	}
	public String getAgentClassString() { return "OTPL" ; }
	public Object report(Argument args[], Context context) 
	    throws ExtensionException, LogoException {
	    return new Interaction(args[0].getString(), args[1].getList(), args[2].getList(), args[3].getList()) ;
	}
    }

    //----------------------------------------------------------------
    public static class InteractionFromList extends DefaultReporter {
	public Syntax getSyntax() {
	    return Syntax.reporterSyntax(new int[] {Syntax.ListType()}, Syntax.WildcardType()) ;
	}
	public String getAgentClassString() { return "OTPL" ; }
	public Object report(Argument args[], Context context) 
	    throws ExtensionException, LogoException {
	    return new Interaction(args[0].getList()) ;
	}
    }    

    //----------------------------------------------------------------
    public static class GetInteractions extends DefaultReporter {
	public Syntax getSyntax() {
	    return Syntax.reporterSyntax(new int[] {}, Syntax.ListType()) ;
	}
	public String getAgentClassString() { return "OTPL" ; }
	public Object report(Argument args[], Context context) 
	    throws ExtensionException, LogoException {
	    LogoListBuilder l = new LogoListBuilder() ;
	    l.addAll(interactions.values()) ;
	    return l.toLogoList() ;
	}
    }    

    //----------------------------------------------------------------
    public static class GetInteraction extends DefaultReporter {
	public Syntax getSyntax() {
	    return Syntax.reporterSyntax(new int[] {Syntax.StringType()}, Syntax.WildcardType()) ;
	}
	public String getAgentClassString() { return "OTPL" ; }
	public Object report(Argument args[], Context context) 
	    throws ExtensionException, LogoException {
	    String name = args[0].getString().toLowerCase() ;
	    if (interactions.containsKey(name))
		return interactions.get(name) ;
	    return false ;
	}
    }    

    //----------------------------------------------------------------
    public static class EvalTrigger extends DefaultReporter {
	public Syntax getSyntax() {
	    return Syntax.reporterSyntax(new int[] {Syntax.WildcardType(), (Syntax.AgentType() | Syntax.NobodyType())}, Syntax.ListType()) ;
	}
	public String getAgentClassString() { return "TP" ; } 
	public Object report(Argument args[], Context context) 
	    throws ExtensionException, LogoException {
	    Object arg0 = args[0].get() ;
	    if (! (arg0 instanceof Interaction))
		throw new ExtensionException("not an interaction: " + org.nlogo.api.Dump.logoObject(arg0)) ;
	    Interaction i = (Interaction) arg0 ;
	    Agent target = null ;
	    try {
		target = (Agent)  args[1].get() ;
	    } catch (ClassCastException e) {}
	    if (! ((target instanceof Turtle) || (target instanceof Patch) || (target == null)))
		throw new ExtensionException("Second argument of eval-trigger should be a turtle or patch or nobody") ;
	    return i.evalReporters(i.getTrigger(), (Agent)context.getAgent(), target) ; 
	}
    }    

    //----------------------------------------------------------------
    public static class EvalCondition extends DefaultReporter {
	public Syntax getSyntax() {
	    return Syntax.reporterSyntax(new int[] {Syntax.WildcardType(), (Syntax.AgentType() | Syntax.NobodyType())}, Syntax.ListType()) ;
	}
	public String getAgentClassString() { return "TP" ; }
	public Object report(Argument args[], Context context) 
	    throws ExtensionException, LogoException {
	    Object arg0 = args[0].get() ;
	    if (! (arg0 instanceof Interaction))
		throw new ExtensionException("not an interaction: " + org.nlogo.api.Dump.logoObject(arg0)) ;
	    Interaction i = (Interaction) arg0 ;
	    Agent target = null ;
	    try {
		target = (Agent)  args[1].get() ;
	    } catch (ClassCastException e) {}
	    if (! ((target instanceof Turtle) || (target instanceof Patch) || (target == null)))
		throw new ExtensionException("Second argument of eval-condition should be a turtle or patch or nobody") ;
	    return i.evalReporters(i.getCondition(), (Agent)context.getAgent(), target) ; 
	}
    }    

    //----------------------------------------------------------------
    public static class PerformActions extends DefaultReporter {
	public Syntax getSyntax() {
	    return Syntax.reporterSyntax(new int[] {Syntax.WildcardType(), (Syntax.AgentType() | Syntax.ListType() | Syntax.NobodyType())}, Syntax.ListType()) ;
	}
	public String getAgentClassString() { return "TP" ; }
	public Object report(Argument args[], Context context) 
	    throws ExtensionException, LogoException {
	    Object arg0 = args[0].get() ;
	    if (! (arg0 instanceof Interaction))
		throw new ExtensionException("not an interaction: " + org.nlogo.api.Dump.logoObject(arg0)) ;
	    Interaction i = (Interaction) arg0 ;
	    Object target = args[1].get() ;
	    if (target != null) {
		if ((target.equals(Nobody.class)) || (target instanceof Nobody) || (org.nlogo.api.Dump.logoObject(target).toString().equals("nobody"))) 
		    target = null ;
		else 
		if (!((target instanceof Turtle) || (target instanceof Patch) 
		      || (target instanceof LogoList)))
		    throw new ExtensionException("Second argument of perform-actions ("+org.nlogo.api.Dump.logoObject(target)+") should be a turtle or patch or a list or nobody") ;
	    }
	    return i.performActions(i.getActions(), (Agent)context.getAgent(), target) ; 
	}
    }    
}
